/*
This program creates the following pipelines using a Java process
as the parent (it doesn't matter if the children processes are written
in Java or C or any other language).
                                                    parent
                                                +------------+
                                                |            |
stdin--+---------------------------------------->>0        1>>--->stdout
       |                                        |            |
       |      child_1                 pipe      |            |
       |   +------------+   +------0========0--->>3        2>>--->stderr
       |   |            |   |                   |            |
       +--->>0        1>>---+   +--0========0--->>4          |
       |   |            |       |     pipe      |            |
       |   |          2>>       |               +------------+
       |   |            |       |           CreateTwoPipedChildren_CompletionService.class
       |   +------------+       |
       |   weird_echo.exe       |
       |                        |
       |                        |
       |      child_2           |
       |   +------------+       |
       |   |            |       |
       +--->>0        1>>-------+
           |            |
           |          2>>
           |            |
           +------------+
           weird_echo.exe

To set up these pipelines, we need to create each child process,
have each child inherit its stdin from the parent, and then
connect each child's stdout to a (new) input stream in the
parent (surprisingly in Java there is no need to explicitly
create "pipe" objects).

Notice how the parent's stdin ends up being connected to all three
processes. This is not ideal. In the Windows C version of this program,
each process has its own independent input stream. I cannot figure out
how to do that with Java.

This program (the parent process) reads from its three input streams
(denoted 0, 3, and 4) in parallel, using FutureTask objects and a
combination of an Executor (to run the tasks) with a CompletionService
(to let us know when each task has completed). Each FutureTask object
that this program creates has the task of reading and filtering one line
from a Scanner object. The tasks are run by an Executor service with a
thread pool of three threads. Instead of polling Future objects (as was done
in both CreateTwoPipedChildren_Executor and CreateTwoPipedChildren_Futures)
this program blocks on an ExecutorCompletionService object which gets
completed tasks from the Executor.
*/

import java.io.IOException;
import java.util.Scanner;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutorCompletionService;

public class CreateTwoPipedChildren_CompletionService
{
   public static void main(String[] args) throws IOException, InterruptedException, ExecutionException
   {
      // Create a command line for running first child.
      ProcessBuilder pb1 = new ProcessBuilder("weird_echo.exe");
    //ProcessBuilder pb1 = new ProcessBuilder("cmd.exe", "/C", "start", "weird_echo.exe");
      // Have the child inherit the parent's standard input stream.
      pb1.redirectInput( ProcessBuilder.Redirect.INHERIT );
      // Execute the child.
      Process process1 = pb1.start();  // this throws IOException


      // Create a command line for running second child.
      ProcessBuilder pb2 = new ProcessBuilder("weird_echo.exe");
    //ProcessBuilder pb2 = new ProcessBuilder("cmd.exe", "/C", "start", "weird_echo.exe");
      // Have the child inherit the parent's standard input stream.
      pb2.redirectInput( ProcessBuilder.Redirect.INHERIT );
      // Execute the child.
      Process process2 = pb2.start();  // this throws IOException


      // Get an input stream that reads from the first child's stdout.
      // This is the first "pipeline" from the first child into the parent.
      Scanner scanner1 = new Scanner( process1.getInputStream() );

      // Get an input stream that reads from the second child's stdout.
      // This is the second "pipeline" from the second child into the parent.
      Scanner scanner2 = new Scanner( process2.getInputStream() );

      // Create a Scanner object to make it easier to use System.in.
      Scanner scanner0 = new Scanner( System.in );


      ExecutorService pool = Executors.newFixedThreadPool(3);
      CompletionService<String> cs = new ExecutorCompletionService<>(pool);

      Task task1 = new Task(scanner1, 2); // double vowels
      Task task2 = new Task(scanner2, 3); // triple vowels
      Task task0 = new Task(scanner0, 4); // quadruple vowels

      Future<String> future0 = cs.submit(task0); // put the three initial tasks into the executor's work queue
      Future<String> future1 = cs.submit(task1);
      Future<String> future2 = cs.submit(task2);

      // The following loop does not do polling as was done in
      //   CreateTwoPipedChildren_Futures.java
      // and
      //   CreateTwoPipedChildren_Executor.java.

      // Each iteration of this loop blocks on a call to the
      // take() method of the ExecutorCompletionService object.
      // The take() method unblocks when a task has been completed
      // by the underlying ExecutorService thread pool. The take()
      // method returns to us a (done) Future object that represents
      // the completed task. We get() the result from the completed
      // Future object, and then create and submit a new Task object
      // to the ExecutorCompletionService, which gives back to us a
      // new Future object representing that new Task.
      int eofs = 0;
      while( eofs < 3 )
      {
         Future<String> future = cs.take(); // wait for a task to complete

         if (future == future0)
         {
            if (null != future0.get()) // not eof
            {
               System.out.println( future0.get() );
               future0 = cs.submit(new Task(scanner0, 4));
            }
            else
            {
               ++eofs;
               System.out.println("Parent ended.");
            }
         }
         else if (future == future1)
         {
            if (null != future1.get()) // not eof
            {
               System.out.println( future1.get() );
               future1 = cs.submit(new Task(scanner1, 2));
            }
            else
            {
               ++eofs;
               System.out.println("Child 1 ended.");
            }
         }
         else if (future == future2)
         {
            if (null != future2.get()) // not eof
            {
               System.out.println( future2.get() );
               future2 = cs.submit(new Task(scanner2, 3));
            }
            else
            {
               ++eofs;
               System.out.println("Child 2 ended.");
            }
         }
      }
      System.out.println( pool.toString() );
      pool.shutdown();
   }
}



// A Task object, when it is called by one of the threads
// in the Executor's thread pool, will read and filter one
// line of input from the Scanner passed to its constructor.
class Task implements Callable<String>
{
   private final Scanner scanner;
   private final int count;

   public Task(final Scanner scanner, final int count)
   {
      this.scanner = scanner;
      this.count = count;
   }

   public String call()
   {
      String result = null;
      // Read one line of data from the Scanner object.
      if ( scanner.hasNextLine() ) // this blocks
      {
         String oneLine = scanner.nextLine();
         StringBuilder temp = new StringBuilder();

         // Filter the line of input before saving it.
         for (int i = 0; i < oneLine.length(); i++)
         {
            temp.append( oneLine.charAt(i) );
            if ( -1 != "aeiouAEIOU".indexOf( oneLine.charAt(i) ) )
            {
               for (int j = 0; j < count - 1; j++)
               {
                  temp.append( oneLine.charAt(i) ); // duplicate vowels
               }
            }
         }
         result = temp.toString();
      }
      else
      {
         result = null; // denote eof
      }
      return result;
   }
}//Task
