/*
This program creates the following pipelines using a Java process
as the parent (it doesn't matter if the children processes are written
in Java or C or any other language).
                                                    parent
                                                +------------+
                                                |            |
stdin--+---------------------------------------->>0        1>>--->stdout
       |                                        |            |
       |      child_1                 pipe      |            |
       |   +------------+   +------0========0--->>3        2>>--->stderr
       |   |            |   |                   |            |
       +--->>0        1>>---+   +--0========0--->>4          |
       |   |            |       |     pipe      |            |
       |   |          2>>       |               +------------+
       |   |            |       |           CreateTwoPipedChildren_Futures.class
       |   +------------+       |
       |   weird_echo.exe       |
       |                        |
       |                        |
       |      child_2           |
       |   +------------+       |
       |   |            |       |
       +--->>0        1>>-------+
           |            |
           |          2>>
           |            |
           +------------+
           weird_echo.exe

To set up these pipelines, we need to create each child process,
have each child inherit its stdin from the parent, and then
connect each child's stdout to a (new) input stream in the
parent (surprisingly in Java there is no need to explicitly
create "pipe" objects).

Notice how the parent's stdin ends up being connected to all three
processes. This is not ideal. In the Windows C version of this program,
each process has its own independent input stream. I cannot figure out
how to do that with Java.

This program (the parent process) reads from its three input streams
(denoted 0, 3, and 4) in parallel, using Future objects. Each Future
object that this program creates has the task of reading and filtering
one line from a Scanner object. Each Future object that this program
creates runs in its own thread, so this program creates a thread for
every line of input that is read. That is not a good design (too many
threads). This will be fixed in CreateTwoPipedChildren_Executor.java.
*/

import java.io.IOException;
import java.util.Scanner;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

public class CreateTwoPipedChildren_Futures
{
   public static void main(String[] args) throws IOException, InterruptedException
   {
      // Create a command line for running first child.
      ProcessBuilder pb1 = new ProcessBuilder("weird_echo.exe");
    //ProcessBuilder pb1 = new ProcessBuilder("cmd.exe", "/C", "start", "weird_echo.exe");
      // Have the child inherit the parent's standard input stream.
      pb1.redirectInput( ProcessBuilder.Redirect.INHERIT );
      // Execute the child.
      Process process1 = pb1.start();  // this throws IOException


      // Create a command line for running second child.
      ProcessBuilder pb2 = new ProcessBuilder("weird_echo.exe");
    //ProcessBuilder pb2 = new ProcessBuilder("cmd.exe", "/C", "start", "weird_echo.exe");
      // Have the child inherit the parent's standard input stream.
      pb2.redirectInput( ProcessBuilder.Redirect.INHERIT );
      // Execute the child.
      Process process2 = pb2.start();  // this throws IOException


      // Get an input stream that reads from the first child's stdout.
      // This is the first "pipeline" from the first child into the parent.
      Scanner scanner1 = new Scanner( process1.getInputStream() );

      // Get an input stream that reads from the second child's stdout.
      // This is the second "pipeline" from the second child into the parent.
      Scanner scanner2 = new Scanner( process2.getInputStream() );

      // Create a Scanner object to make it easier to use System.in.
      Scanner scanner0 = new Scanner( System.in );


      FutureLine future1 = new FutureLine(scanner1, 2); // double vowels
      FutureLine future2 = new FutureLine(scanner2, 3); // triple vowels
      FutureLine future0 = new FutureLine(scanner0, 4); // quadruple vowels

      boolean child1EOF = false;
      boolean child2EOF = false;
      boolean parentEOF = false;

      // The following loop polls the three FutureLine objects to see
      // when one of them is done. When a FutureLine object has finished
      // its task, the Future's value is printed out and a new FutureLine
      // object is created to read the next line from the appropriate
      // input stream. Each FutureLine runs in its own thread. Notice that
      // one thread is created for every line of input read from each
      // of the three input streams. Also notice that the main() thread
      // is the thread that does the polling and the instantiating of
      // new FutureLine objects.
      while( ! parentEOF || ! child1EOF || ! child2EOF )
      {
         if( ! parentEOF && future0.isDone() )
         {
            if (null != future0.get()) // not eof
            {
               System.out.println( future0.get() );
               future0 = new FutureLine(scanner0, 4); // quadruple vowels
            }
            else
            {
               parentEOF = true;
               System.out.println("Parent ended.");
            }
         }
         if( ! child1EOF && future1.isDone() )
         {
            if (null != future1.get()) // not eof
            {
               System.out.println( future1.get() );
               future1 = new FutureLine(scanner1, 2); // double vowels
            }
            else
            {
               child1EOF = true;
               System.out.println("Child 1 ended.");
            }
         }
         if( ! child2EOF && future2.isDone() )
         {
            if (null != future2.get()) // not eof
            {
               System.out.println( future2.get() );
               future2 = new FutureLine(scanner2, 3); // triple vowels
            }
            else
            {
               child2EOF = true;
               System.out.println("Child 2 ended.");
            }
         }
      }
   }
}



// A FutureLine object will read and filter one line
// of input from the Scanner passed to its constructor.
class FutureLine implements Future<String>
{
   public String result;
   Thread thread = null;
   boolean cancelled = false;

   public FutureLine(final Scanner scanner, final int count)
   {
      this.thread = new Thread(){
         public void run()
         {
            // Read one line of data from the Scanner object.
            if ( scanner.hasNextLine() ) // this blocks
            {
               String oneLine = scanner.nextLine();
               StringBuilder temp = new StringBuilder();

               // Filter the line of input before saving it.
               for (int i = 0; i < oneLine.length(); i++)
               {
                  temp.append( oneLine.charAt(i) );
                  if ( -1 != "aeiouAEIOU".indexOf( oneLine.charAt(i) ) )
                  {
                     for (int j = 0; j < count - 1; j++)
                     {
                        temp.append( oneLine.charAt(i) ); // duplicate vowels
                     }
                  }
               }
               result = temp.toString();
            }
            else
            {
               result = null; // denote eof
            }
         }
      };
      thread.start();
   }

   public String get()
   {
      // Waits if necessary for the computation to complete, and then retrieves its result.
      try
      {
         this.thread.join();
      }
      catch (InterruptedException e){System.err.println(e);}
      return this.result;
   }

   public String get(long timeout, TimeUnit unit) throws TimeoutException
   {
      //Waits if necessary for at most the given time for the computation to complete,
      // and then retrieves its result, if available.
      try
      {
         this.thread.join(timeout);
      }
      catch (InterruptedException e){System.err.println(e);}
      if ( thread.isAlive() )
      {
         throw new TimeoutException("future result not ready");
      }
      else
      {
         return this.result;
      }
   }

   public boolean isDone()
   {
      // Returns true if this task completed.
      return ! thread.isAlive();
   }

   public boolean cancel(boolean mayInterruptIfRunning)
   {
      // Attempts to cancel execution of this task.
      if ( isDone() || isCancelled() )
      {
         return false;
      }
      else
      {
         thread.interrupt();
         cancelled = true;
         return cancelled;
      }
   }

   public boolean isCancelled()
   {
      // Returns true if this task was canceled before it completed normally.

      return cancelled;
   }
}//FutureLine
