/*
This program creates the following pipelines using a Java process
as the parent (it doesn't matter if the children processes are written
in Java or C or any other language).
                                                    parent
                                                +------------+
                                                |            |
stdin--+---------------------------------------->>0        1>>--->stdout
       |                                        |            |
       |      child_1                 pipe      |            |
       |   +------------+   +------0========0--->>3        2>>--->stderr
       |   |            |   |                   |            |
       +--->>0        1>>---+   +--0========0--->>4          |
       |   |            |       |     pipe      |            |
       |   |          2>>       |               +------------+
       |   |            |       |           CreateTwoPipedChildren_CompletionService.class
       |   +------------+       |
       |   weird_echo.exe       |
       |                        |
       |                        |
       |      child_2           |
       |   +------------+       |
       |   |            |       |
       +--->>0        1>>-------+
           |            |
           |          2>>
           |            |
           +------------+
           weird_echo.exe

To set up these pipelines, we need to create each child process,
have each child inherit its stdin from the parent, and then
connect each child's stdout to a (new) input stream in the
parent (surprisingly in Java there is no need to explicitly
create "pipe" objects).

Notice how the parent's stdin ends up being connected to all three
processes. This is not ideal. In the Windows C version of this program,
each process has its own independent input stream. I cannot figure out
how to do that with Java.

This program (the parent process) reads from its three input streams
(denoted 0, 3, and 4) in parallel. The work of reading and filtering
a line of text from each input stream is split up into two tasks, an
I/O task and a compute task. This program has two thread pools, one
for each kind of task.

The I/O tasks are done in this program using FutureTask objects and a
combination of an Executor (to run the I/O tasks) with a CompletionService
(to let us know when each I/O task has completed). Each FutureTask object
has the task of reading just one line from a Scanner object. These I/O
tasks are run by an Executor service with a thread pool of three threads.
This program blocks on a ExecutorCompletionService object which gets
completed I/O tasks from the Executor.

When an I/O task completes, and is returned to us by the take() method form
the ExecutorCompletionService, we create a compute task and submit it to the
other thread pool. Each compute task is represented by a Runnable object that
filters and prints one line of text.
*/

import java.io.IOException;
import java.util.Scanner;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutorCompletionService;

public class CreateTwoPipedChildren_Pipelined
{
   public static void main(String[] args) throws IOException, InterruptedException, ExecutionException
   {
      // Create a command line for running first child.
      ProcessBuilder pb1 = new ProcessBuilder("weird_echo.exe");
    //ProcessBuilder pb1 = new ProcessBuilder("cmd.exe", "/C", "start", "weird_echo.exe");
      // Have the child inherit the parent's standard input stream.
      pb1.redirectInput( ProcessBuilder.Redirect.INHERIT );
      // Execute the child.
      Process process1 = pb1.start();  // this throws IOException


      // Create a command line for running second child.
      ProcessBuilder pb2 = new ProcessBuilder("weird_echo.exe");
    //ProcessBuilder pb2 = new ProcessBuilder("cmd.exe", "/C", "start", "weird_echo.exe");
      // Have the child inherit the parent's standard input stream.
      pb2.redirectInput( ProcessBuilder.Redirect.INHERIT );
      // Execute the child.
      Process process2 = pb2.start();  // this throws IOException


      // Get an input stream that reads from the first child's stdout.
      // This is the first "pipeline" from the first child into the parent.
      Scanner scanner1 = new Scanner( process1.getInputStream() );

      // Get an input stream that reads from the second child's stdout.
      // This is the second "pipeline" from the second child into the parent.
      Scanner scanner2 = new Scanner( process2.getInputStream() );

      // Create a Scanner object to make it easier to use System.in.
      Scanner scanner0 = new Scanner( System.in );


      // Two thread pools, one for IO operations and one for cpu operations.
      ExecutorService pool_1 = Executors.newFixedThreadPool(3);
      CompletionService<String> cs = new ExecutorCompletionService<>(pool_1);
      ExecutorService pool_2 = Executors.newFixedThreadPool(3);

      Task_1 task1 = new Task_1(scanner1);
      Task_1 task2 = new Task_1(scanner2);
      Task_1 task0 = new Task_1(scanner0);

      Future<String> future0 = cs.submit(task0); // put three initial I/O tasks into the executor's work queue
      Future<String> future1 = cs.submit(task1);
      Future<String> future2 = cs.submit(task2);

      // Each iteration of this loop blocks on a call to the take()
      // method of the ExecutorCompletionService object. The take()
      // method unblocks when an I/O task has been completed by the
      // underlying ExecutorService thread pool. The take() method
      // returns to us a (done) Future object that represents the
      // completed I/O task. We get() the result from the completed
      // Future object and use it to create and submit a new compute
      // task to the second Executor thread pool. We also create and
      // submit a new I/O task object to the ExecutorCompletionService,
      // which gives back to us a new Future object representing that
      // new I/O task.

      int eofs = 0;
      while( eofs < 3 )
      {
         Future<String> future = cs.take(); // wait for a task to complete

         if (future == future0)
         {
            if (null != future0.get()) // not eof
            {
               pool_2.execute( new Task_2(future0.get(), 4) ); // quadruple vowels
               future0 = cs.submit( new Task_1(scanner0) );
            }
            else
            {
               eofs++;
               System.out.println("Parent ended.");
            }
         }
         else if (future == future1)
         {
            if (null != future1.get()) // not eof
            {
               pool_2.execute( new Task_2(future1.get(), 2) ); // double vowels
               future1 = cs.submit( new Task_1(scanner1) );
            }
            else
            {
               eofs++;
               System.out.println("Child 1 ended.");
            }
         }
         else if (future == future2)
         {
            if (null != future2.get()) // not eof
            {
               pool_2.execute( new Task_2(future2.get(), 3) ); // triple vowels
               future2 = cs.submit( new Task_1(scanner2) );
            }
            else
            {
               eofs++;
               System.out.println("Child 2 ended.");
            }
         }
      }
      System.out.println( pool_1.toString() );
      System.out.println( pool_2.toString() );
      pool_1.shutdown();
      pool_2.shutdown();
   }
}



// A Task_1 object, when it is called by one of the threads
// in the Executor's thread pool, will read one line of
// input from the Scanner passed to its constructor.
class Task_1 implements Callable<String>
{
   private final Scanner scanner;

   public Task_1(final Scanner scanner)
   {
      this.scanner = scanner;
   }

   public String call()
   {
      String result = null;
      // Read one line of data from the Scanner object.
      if ( scanner.hasNextLine() ) // this blocks
      {
         result = scanner.nextLine();
         //System.out.println( result ); // combine this with the sleep in Task_2
      }
      else
      {
         result = null; // denote eof
      }
      return result;
   }
}//Task_1


// A Task_2 object, when it is called by one of the threads in
// the Executor's thread pool, will filter one line of text.
class Task_2 implements Runnable
{
   private final String oneLine;
   private final int count;

   public Task_2(final String oneLine, final int count)
   {
      this.oneLine = oneLine;
      this.count = count;
   }

   public void run()
   {
      StringBuilder temp = new StringBuilder();

      //try{Thread.sleep(5000);}catch(Exception e){} // combine this with the println in Task_1

      // Filter the line of input.
      for (int i = 0; i < oneLine.length(); i++)
      {
         temp.append( oneLine.charAt(i) );
         if ( -1 != "aeiouAEIOU".indexOf( oneLine.charAt(i) ) )
         {
            for (int j = 0; j < count - 1; j++)
            {
               temp.append( oneLine.charAt(i) ); // duplicate vowels
            }
         }
      }
      System.out.println( temp );
   }
}//Task_2
